% loesungen.m
% Kapitel 6: Signalverarbeitung - Signal Processing Toolbox
%            Lsungen zu den bungsaufgaben

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% 6.5.1 Signaltransformation im Frequenzbereich

t = 0:0.001:0.049;                    % Zeitvektor   1000 Hz
x = square (2*pi*20*t);               % Rechtecksignal 20 Hz

y = fft (x);                          % DFT (ohne Normierung)
n = length (y);                       % Lnge des Datensatzes
f = (0:n/2) / n / diff(t(1:2));       % Frequenzvektor zum Plotten

                                      % Erzeugt Skalierungsvektor
skal = imag (exp (j*pi/2*(0:n/2)));   % 0 1 0 -1   0 1   0 -1   ...
skal = skal ./ [1 (1:n/2)];           % 0 1 0 -1/3 0 1/5 0 -1/7 ...
skal = [skal skal(n/2:-1:2)];         % Spiegeln fr Frequenzen > 500 Hz

y_skal = y .* skal;                   % Skalierte Fourierkoeffizienten
x_skal = ifft (y_skal);               % inverse DFT (ohne Normierung)

figure (1)
subplot (121)
    plot (t, x, 'b--')
    hold on
    plot (t, real (x_skal), 'r-')
    ylim ([-1.7 1.7])
    legend ('Rechteck', 'Dreieck', 'Location', 'NE')
    title ('Zeitsignale')
    xlabel ('Zeit [s]')
    
subplot (122)
    stem (f, -2*imag (y(1:n/2+1))/n, 'bx')
    hold on
    stem (f, -2*imag (y_skal(1:n/2+1))/n, 'ro')
    ylim ([-0.3 1.5])
    legend ('Rechteck', 'Dreieck', 'Location', 'NE')
    title ('Fourierkoeffizienten b_k')
    xlabel ('Frequenz [Hz]')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% 6.5.2 Signalanalyse und digitale Filterung

% Der Datensatz wurde mit den folgenden Zeilen erzeugt und gespeichert:

t = 0.001:0.001:1;                                   % Zeitvektor Fs = 1000 Hz
xx = 12 * sin (2*pi*4*t) + ...                       % Sollsignal      4 Hz
      4 * sin (2*pi*12*t);                           %                12 Hz
x =  xx + ...
     20 * cos (2*pi*63*t) + ...                      % Strsignal     63 Hz
     18 * cos (2*pi*137*t) .* sin (2*pi*29*t) +  ... % 137-29 Hz, 137+29 Hz
      4 + randn (size(t));                           % Rauschen

datei = fopen ('ueb_signal.dat', 'w');               % Signal speichern (ASCII)
fprintf (datei, '%5.3f  %6.2f\r\n', [t' x']');       % \r\n = CRLF im DOS-Format
fclose (datei);

save ueb_signal t x                                  % Signal speichern (MAT)

% Hinweis: Die Multiplikation eines 137-Hz mit einem 29-Hz-Signal
% (d.h. Amplitudenmodulation) entspricht einer Faltung im Frequenzbereich.
% Dadurch werden Anteile mit 108 und 166 Hz erzeugt (=137 +/- 29 Hz).

clear

% Lsung der Aufgabe

load ueb_signal.dat                          % Signal laden
t = ueb_signal (:,1)';                       % Zeitvektor
x = ueb_signal (:,2)';                       % Messdaten-Vektor

T = diff (t(1:2));                           % Abtastzeit
N = length (x);                              % Lnge des Datenvektors
f = [0:floor((N-1)/2)] / (N*T);              % Frequenzvektor fr Plot

X = fft (x);                                 % Fouriertransformation
X = [X(1) 2*X(2:floor((N-1)/2)+1)] / N;      % Begrenzung und Normierung

                                             % FIR-Filter auslegen
ord    = 40;                                 % Ordnung des FIR-Filters
B      = fir1 (ord, 0.04);                   % FIR-Filter berechnen f_g = 20 Hz
x_filt = filter (B, 1, x);                   % Daten filtern
x_filt = x_filt - mean (x_filt);             % Gleichanteil ausblenden

figure
subplot (221)
    plot (t, x, 'b-')
    xlabel ('Zeit [s]')
    title ('Ungefiltertes Signal')
subplot (222)
    stem (f, abs (X), 'bo')
    axis ([0 500 0 25])
    xlabel ('Frequenz [Hz]')
    title ('Spektrum')
subplot (223)
    plot (t, x_filt, 'r-')
    xlabel ('Zeit [s]')
    title ('Gefiltertes Signal')

% Die bertragungsfunktion des FIR-Filters kann angezeigt werden mit:

F   = [-ord/2:0.1:ord/2] / (ord*T);
amp = zeros (size (F));
for i = -ord/2:ord/2
   amp = amp + B(i+ord/2+1) * cos (2*pi*F*i*T);
end
subplot (224)
    plot (F, abs (amp))
    axis ([0 500 0 1])
    xlabel ('Frequenz [Hz]')
    title ('Filter-bertragungsfunktion')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% 6.5.3 Analoger Bandpass

% Es werden ein Bandpass (10...100 Hz) und eine Bandsperre (20...70 Hz)
% erzeugt. Die Faltung beider bertragungsfunktionen ergibt das gewnschte Filter.

[B1, A1] = butter (8, 2*pi*[10 100], 's');        % Bandpass 10...100 Hz
[B2, A2] = butter (8, 2*pi*[20 70], 'stop', 's'); % Bandsperre 20...70 Hz

A = conv (A1, A2);                                 % Nenner Falten
B = conv (B1, B2);                                 % Zhler Falten

W = 2*pi * linspace (0, 150, 301);                 % Frequenzvektor [rad/s]
H = freqs (B, A, W);                               % Gesamt-bertragungsfunktion

figure
    plot ([0 10 10 20 20 70 70 100 100 150], ...       % ideale bertragungs-
          [0  0  1  1  0  0  1   1   0   0], 'k--')    % funktion
    hold on
    plot (W/2/pi, abs (H), 'b-')
    axis ([0 150 0 1.2])
    xlabel ('Frequenz [Hz]')
    title ('Gesamt-Filter')
    legend ('Ideal', '8. Ord.')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% 6.5.4 Digitaler IIR-Bandpass

% Die einfachste Berechnung erfolgt mit dem Befehl BUTTER analog zur
% vorherigen Aufgabe. Die Grenzfrequenzen sind jetzt normiert;
% FREQZ bentigt zustzlich die Abtastfrequenz Fs.

Fs = 300;                                          % Abtastfrequenz 300 Hz
[B1, A1] = butter (8, [10 100]/(Fs/2));            % Bandpass 10...100 Hz
[B2, A2] = butter (8, [20 70]/(Fs/2), 'stop');     % Bandsperre 20...70 Hz

A = conv (A1, A2);                                 % Nenner Falten
B = conv (B1, B2);                                 % Zhler Falten

F = linspace (0, 150, 301);                        % Frequenzvektor [Hz]
H = freqz (B, A, F, Fs);                           % Gesamt-bertragungsfunktion

    plot (F, abs (H), 'g-')

% Die Berechnung mittels Prototyp-Tiefpass fhrt zum selben Ergebnis,
% erfordert aber ein individuelles Pre-warping fr jede Grenzfrequenz.
% Die bertragungsfunktion des Filters im Laplace-Bereich kann nicht
% direkt transformiert werden, da dann das Pre-warping nur fr eine
% Frequenz optimal passen wrde.

% Liegt die Abtastfrequenz deutlich ber der hchsten Grenzfrequenz, nimmt die
% Bedeutung des Pre-warping aber ab und die direkte Tranfsormation liefert
% eine gute Nherung; allerdings kann es dann zu numerischen Problemen
% aufgrund der starken Grenunterschiede der Koeffizienten geben.

[z, p, k]    = buttap (8);                         % Prototyp-Tiefpass
[a, b, c, d] = zp2ss (z, p, k);                    % Umrechnung in Zustandsform

omega        = [10 100] * 2 * pi;                  % Grenzfrequenz
omega        = 2 * Fs * tan (omega / Fs / 2);      % Grenzfrequenz mit Pre-warping
[e, f, g, h] = lp2bp (a, b, c, d, sqrt(prod(omega)), diff(omega));   % Bandpass
[r, s, t, u] = bilinear (e, f, g, h, Fs);          % Dig. Filter mit Pre-warping
[B1, A1]     = ss2tf (r, s, t, u);                 % bertragungsfunktion

omega        = [20 70] * 2 * pi;                   % Grenzfrequenz
omega        = 2 * Fs * tan (omega / Fs / 2);      % Grenzfrequenz mit Pre-warping
[e, f, g, h] = lp2bs (a, b, c, d, sqrt(prod(omega)), diff(omega));   % Bandsperre
[r, s, t, u] = bilinear (e, f, g, h, Fs);          % Dig. Filter mit Pre-warping
[B2, A2]     = ss2tf (r, s, t, u);                 % bertragungsfunktion

A = conv (A1, A2);                                 % Nenner Falten
B = conv (B1, B2);                                 % Zhler Falten

F = linspace (0, 150, 301);                        % Frequenzvektor [Hz]
H = freqz (B, A, F, Fs);                           % Gesamt-bertragungsfunktion

    plot (F, abs (H), 'c--')
 
% Der Befehl YULEWALK erzeugt ein Filter mit hherer Flankensteilheit aber
% auch deutlich hherer Welligkeit sowohl in den Durchlass- als auch in den
% Sperrbereichen.

frequenz  = [0 10 10 20 20 70 70 100 100 150] / (Fs/2); % Normierte Frequenzen
amplitude = [0  0  1  1  0  0  1   1   0   0];     % Amplituden
[B, A] = yulewalk (16, frequenz, amplitude);       % IIR-Filter mit Ordnung 2*8

F = linspace (0, 150, 301);                        % Frequenzvektor [Hz]
H = freqz (B, A, F, Fs);                           % Gesamt-bertragungsfunktion

    plot (F, abs (H), 'r-')
    axis ([0 150 0 1.2])
    xlabel ('Frequenz [Hz]')
    title ('Gesamt-Filter')
    legend ('Ideal', 'Analog', 'IIR', 'IIR Prototyp', 'YULEWALK')
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
